package ensembles;

import java.util.Collection;
import java.util.Iterator;

import listes.Liste;

/**
 * Ensemble trié utilisant un {@link Liste}
 *
 * @author davidroussel
 */
public class EnsembleTriListe<E extends Comparable<E>> extends
		EnsembleListe<E> implements EnsembleTri<E>
{

	/**
	 * Constructeur par défaut d'un ensemble trié utilisant un {@link Liste}
	 */
	public EnsembleTriListe()
	{
		liste = new Liste<E>();
	}

	/**
	 * Constructeur de copie à partir d'un autre iterable
	 *
	 * @param elements l'itérable dont on veut copier les éléments
	 */
	public EnsembleTriListe(Iterable<E> elements)
	{
		liste = new Liste<E>();
		for (E elt : elements)
			ajout(elt);
	}

	/**
	 * Ajout d'un élément de manière triée dans l'ensemble utilisant un
	 * {@link Liste}
	 * @param element l'élément à ajouter de manière triée (on considèrera que
	 *            l'on ne peut pas ajouter d'élément null)
	 * @return true si l'élément n'était pas déjà présent dans l'ensemble, false
	 *         sinon ou si l'on a tenté d'insérer un élément null (auquel cas il
	 *         n'est pas inséré).
	 * @see ensembles.EnsembleListe#ajout(java.lang.Object)
	 * @see listex.Liste#insertElement(E, int)
	 */
	@Override
	public boolean ajout(E element)
	{
		if (element == null) return false;
		if (contient(element)) return false;
		
		liste.insert(element, rang(element));
		
		return true;
	}

	/**
	 * Test d'égalité d'un ensemble trié. Il est nécessaire de réimplémenter la
	 * comparaison avec un autre ensemble car l'ordre des élément aura son
	 * importance dans la comparaison ce qui n'était pas le cas avec les
	 * ensembles non triés.
	 *
	 * @return true si l'objet obj est aussi un ensemble (pas forcément trié) et
	 *         qu'il contient exactement les mêmes éléments dans le même ordre.
	 * @see ensembles.EnsembleGenerique#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj)
	{
		/* DONE
		 *  Remplacer par ...
		 * 	1 - obj == null ? ==> false
		 * 	2 - obj == this ? ==> true
		 * 	3 - obj est une instance de Ensemble<?>
		 * 		- caster obj en Ensemble<?>
		 * 		- si obj et this ont exactement les mêmes éléments dans le
		 * 		  même ordre ==> true
		 * 		- sinon ==> false;
		 * 	  - sinon (obj n'est pas un Ensembble<?>) ==> false
		 */
		if (obj == null) return false;
		if (obj == this) return true;
		if (obj instanceof Ensemble<?>)
		{
			Ensemble<?> objE = (Ensemble<?>) obj;
			if (cardinal() != objE.cardinal()) return false;
			
			Iterator<?> itObj = objE.iterator();
			Iterator<?> itThis = iterator();
			while (itObj.hasNext())	//pas besoin de tester les deux, on a déjà vérifié qu'ils avaient le même nombre d'éléments.
				if (itObj.next() != itThis.next())
					return false;
		}
		else return false;
		
		return true;
	}

	/**
	 * Code de hachage d'un ensemble trié. Il est nécessaire de réimplémenter le
	 * code de hachage pour les ensembles triés car on considèrera que deux
	 * ensembles contenant les même éléments mais dans des ordres différents
	 * seront eux-même différents. Il faut donc que la méthode hashCode prenne
	 * en compte l'ordre des éléments (Comme dans les autres {@link Collection}
	 * d'ailleurs).
	 *
	 * @return le code de hachage de cet ensemble trié.
	 * @see listex.Liste#hashCode() pour un exemple de hashage utilisant
	 *      l'ordre des éléments
	 * @see ensembles.EnsembleGenerique#hashCode()
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		for (E elt : this)
			result = result*prime + elt.hashCode();
		
		return result;
	}
}
